#define SIZE 256

#define WHITE 0xFFFFFFFF
#define DIMWHITE 0xFFCFCFCF
#define DIMMERWHITE 0xFF9F9F9F
#define RED   0xFF0000FF

#define SNOW WHITE
#define GREENLAND 0xFF47b61a
#define BROWNLAND 0xFF359db8
#define WATER 0xFFed422b
#define SPACE 0xFF0F0007

union color
{
  struct
  {
    unsigned char r;
    unsigned char g;
    unsigned char b;
    unsigned char a;
  };
  struct
  {
    unsigned char h;
    unsigned char s;
    unsigned char v;
  };
  unsigned int c;
} color[SIZE][SIZE] = { 0 };

typedef struct ivec2
{
  int x;
  int y;
} ivec2;
typedef struct vec2
{
  float x;
  float y;
} vec2;
typedef struct vec3
{
  float x;
  float y;
  float z;
} vec4;

struct rgb
{
  union
  {
    float r;
    float h;
  };
  union
  {
    float g;
    float s;
  };
  union
  {
    float b;
    float v;
  };
};

static inline struct rgb HSVtoRGB(struct rgb hsv)
{
    int i = (int)(hsv.h * 6.0);
    float f = (hsv.h * 6.0) - (float)i;
    float p = hsv.v * (1.0 - hsv.s);
    float q = hsv.v * (1.0 - (f * hsv.s));
    float t = hsv.v * (1.0 - ((1.0 - f) * hsv.s));

    switch(i % 6)
    {
        case 0: return (struct rgb) { .r = hsv.v, .g = t, .b = p }; break;
        case 1: return (struct rgb) { .r = q, .g = hsv.v, .b = p }; break;
        case 2: return (struct rgb) { .r = p, .g = hsv.v, .b = t }; break;
        case 3: return (struct rgb) { .r = p, .g = q, .b = hsv.v }; break;
        case 4: return (struct rgb) { .r = t, .g = p, .b = hsv.v }; break;
        case 5: return (struct rgb) { .r = hsv.v, .g = p, .b = q }; break;
    }
    return (struct rgb) { .r = 0.0, .g = 0.0, .b = 0.0 };
}

#define BODIES 1
struct body
{
  struct vec3 pos;
  float r;
} body[BODIES] =
{
  (struct body)
  {
    .pos.x = 0.0,
    .pos.y = 0.0,
    .pos.z = 0.5,
    .r = 64,
  },
};

#define STARS 125
struct star
{
  struct ivec2 pos;
  unsigned char r;
  float t;
  float n;
} star[STARS];

static unsigned char initialized = 0;

static int rseed = 253918727;

static inline unsigned char rand(void) {
    rseed = (unsigned int){ (214013*rseed+253097717) };
    return (unsigned char){(rseed>>16)};
}

static inline void generateStars(void)
{
  for (int s = 0;s < STARS;s++)
  {
    star[s].pos.x = rand();
    star[s].pos.y = rand();
    unsigned char r = rand() % 6;
    if (r == 5)
      r = rand() % 6;
    star[s].r = ((rand() % (r == 0 ? 1 : r)) << 4) + r;
    star[s].t = 0.1 + (0.1 * ((float)rand() / 256.0));
  }
}

static inline int abs(int a){ return a > 0 ? a : -a; }

static inline void DDA(int X0, int Y0, int X1, int Y1)
{
    // calculate dx & dy
    int dx = X1 - X0;
    int dy = Y1 - Y0;

    // calculate steps required for generating pixels
    int steps = abs(dx) > abs(dy) ? abs(dx) : abs(dy);

    // calculate increment in x & y for each steps
    float Xinc = dx / (float)steps;
    float Yinc = dy / (float)steps;

    // Put pixel for each step
    float X = X0;
    float Y = Y0;
    for (int i = 0; i <= steps; i++)
    {
      color[(int)X][(int)Y].c = 0xFFFFFFFF;
      X += Xinc;
      Y += Yinc;
    }
}

/*
0 = snow
1 = water
2 = green land
3 = brown land
*/
#define WIDTH 32
#define HEIGHT 16
unsigned int image[HEIGHT][2] =
{
  { 0x00000000, 0x00000000, },
  { 0x05555A96, 0xA03FFC02, },
  { 0x95555555, 0x6A16555A, },
  { 0xA9555555, 0x6A9555AA, },
  { 0xAA5FDF6A, 0x6AA57DAA, },
  { 0xAA95FFBF, 0xEAA976AA, },
  { 0xAAA55FFF, 0xFAAAFAAA, },
  { 0xAA969ED7, 0xEA55EAAA, },
  { 0xAA5ABA56, 0xA957AAAA, },
  { 0xA9FAAA96, 0xA95EAAAA, },
  { 0xA7FAAABE, 0xAA5EAAAA, },
  { 0xA7EAAAAA, 0xAABEAAAA, },
  { 0xAAAAAAAA, 0xAAAEAAAA, },
  { 0xAAAAAAAA, 0xAAAAAAAA, },
  { 0x80002000, 0x2AA0202A, },
  { 0x00000000, 0x00000000, },
};

float width[HEIGHT] =
{
  2.0,
  4.0,
  5.0,
  6.0,
  6.0,
  7.0,
  7.0,
  7.0,
  7.0,
  7.0,
  7.0,
  6.0,
  6.0,
  5.0,
  4.0,
  2.0,
};

static inline unsigned int getColor(float x, float y, float r, float a)
{
  y *= a;
  int iy = (y * 8) + 8;
  x *= (7.0 / (width[iy])) * a;
  x += r;
  int ix = (x * 8) + 16;
  int c;
  if (iy < 0 || iy >= HEIGHT)
  {
    c = 0;
  }
  else
  {
    ix = (ix % 32) * 2;
    if (ix >= WIDTH)
    {
      ix -= 32;
      c = (image[iy][1] >> ix) & 0x3;
    }
    else
      c = (image[iy][0] >> ix) & 0x3;
  }
  switch (c)
  {
    case 0: //snow
      return SNOW;
    case 1: //green land
      return GREENLAND;
    case 2: //brown land
      return WATER;
    case 3: //water
      return BROWNLAND;
  }
  return 0x00000000;
}

int game(float time);
int game(float time)
{
  if (initialized == 0)
  {
    generateStars();
    initialized = 1;
  }
  for (int y = 0;y < SIZE;y++)
    for (int x = 0;x < SIZE;x++)
      color[y][x].c = SPACE;

  body[0].pos.z = time/35.0 + 0.3;
  if (body[0].pos.z > 1.0)
    body[0].pos.z = 1.0;

  for (int s = 0;s < STARS;s++)
  {
    star[s].pos.x++;
    star[s].pos.x %= SIZE;
    if (time > star[s].n)
    {
      star[s].n = time + star[s].t;
      if (star[s].r >> 4 == (star[s].r & 0x0F))
        star[s].r &= 0x0F;
      else
        star[s].r += 0x10;
    }
    switch (star[s].r >> 4)
    {
      case 5:
        color[star[s].pos.y][(star[s].pos.x+4)&0xFF].c = DIMMERWHITE;
        color[star[s].pos.y][(star[s].pos.x-4)&0xFF].c = DIMMERWHITE;
        color[(star[s].pos.y+4)&0xFF][star[s].pos.x].c = DIMMERWHITE;
        color[(star[s].pos.y-4)&0xFF][star[s].pos.x].c = DIMMERWHITE;
      case 4:
        color[star[s].pos.y][(star[s].pos.x+2)&0xFF].c = DIMWHITE;
        color[star[s].pos.y][(star[s].pos.x-2)&0xFF].c = DIMWHITE;
        color[(star[s].pos.y+2)&0xFF][star[s].pos.x].c = DIMWHITE;
        color[(star[s].pos.y-2)&0xFF][star[s].pos.x].c = DIMWHITE;
      case 3:
        color[(star[s].pos.y+1)&0xFF][star[s].pos.x].c = WHITE;
        color[(star[s].pos.y-1)&0xFF][star[s].pos.x].c = WHITE;
        color[star[s].pos.y][(star[s].pos.x+1)&0xFF].c = WHITE;
        color[star[s].pos.y][(star[s].pos.x-1)&0xFF].c = WHITE;
        color[star[s].pos.y][star[s].pos.x].c = WHITE;
      case 2:
        color[star[s].pos.y][star[s].pos.x].c = WHITE;
        break;
      case 1:
        color[star[s].pos.y][star[s].pos.x].c = DIMWHITE;
        break;
      case 0:
        color[star[s].pos.y][star[s].pos.x].c = DIMMERWHITE;
        break;
    }
  }

  for (int b = 0;b < BODIES;b++)
  {
    struct origin
    {
      int x;
      int y;
    } origin = (struct origin)
    {
      .x = (body[b].pos.x * body[b].pos.z) + (SIZE/2),
      .y = (body[b].pos.y * body[b].pos.z) + (SIZE/2)
    };
    int radius = body[b].r / body[b].pos.z;
    if (radius > SIZE*2)
      radius = SIZE*2;
    float z = (body[b].pos.z * 8.0 > 4.0 ? 4.0 : body[b].pos.z * 8.0) * 0.25;
    for (int y=-radius; y<=radius; y++)
    {
      for (int x=-radius; x<=radius; x++)
      {
        if (x*x+y*y <= radius*radius)
        {
          if (origin.x+x >= 0 && origin.x+x < SIZE &&
              origin.y+y >= 0 && origin.y+y < SIZE)
          {
            float bx = ((float)x/(float)radius);
            float by = ((float)y/(float)radius);
            color[origin.y+y][origin.x+x].c = getColor(bx,by,time/4.0,z);
          }
        }
      }
    }
  }

  return 0;
}
